/*
	Blender local illumination models
*/

#include "bmisc.h"

color Blnd_LILambert(normal Nn)
{
  color C = 0;
  extern point P;

  illuminance(P, Nn, PI/2) {
	C += Cl * (Nn . normalize(L));
  }

  return C;
}

color Blnd_LIPhong(normal N; vector V; float hard)
{
	color C = 0;
	extern point P;

	illuminance (P, N, PI/2) {
		vector H = normalize(normalize(L) + V);
		C += Cl * pow(max(0, N.H), hard);
	}

	return C;
}

color Blnd_LICookTorr(normal N; vector V; float hard)
{
	color C = 0;
	extern point P;

	illuminance (P, N, PI/2) {
		vector H = normalize(normalize(L) + V);
		float nh = N . H;
		if (nh > 0)
		{
			float nv = max(0, N.V);
			C += Cl * pow(max(0, nh), hard) / (0.1 + nv);
		}
	}

	return C;
}

color Blnd_LIBlinn(normal N; vector V; float hard, ior)
{
	color C = 0;
	extern point P;

	float specPower = (hard < 100) ? sqrt(1 / hard) : (10 / hard);

	illuminance (P, N, PI/2) {
		vector H = normalize(normalize(L) + V);
		float nh = N . H;
		if (nh > 0)
		{
			float nv = N.V;
			if (nv < 0) nv = 0.01;

			float nl = N.L;
			if (nl > 0)
			{
				float vh = V.H;
				if (vh < 0) vh = 0.01;

				float a = 1.0;
				float b = (2.0*nh*nv)/vh;
				float c = (2.0*nh*nl)/vh;
				float g = 0, p, f, ang;

				if( a < b && a < c ) g = a;
				else if( b < a && b < c ) g = b;
				else if( c < a && c < b ) g = c;

				p = sqrt((ior * ior)+(vh*vh)-1.0);
				f = (((p-vh)*(p-vh))/((p+vh)*(p+vh)))*(1+((((vh*(p+vh))-1.0)*((vh*(p+vh))-1.0))/(((vh*(p-vh))+1.0)*((vh*(p-vh))+1.0))));
				ang = saacos(nh);

				float i = max(0, f * g * exp(-ang*ang /
					(2.0*specPower*specPower)));

				C += Cl * i;
			}
		}
	}

	return C;
}
