/*
	Blender textures for Renderman
	v0.2
*/

#include "bwaves.h"
#include "bnoise.h"

normal norm_derivate(float tin; normal N; string coba)
{
  normal n = N;

	if (coba != "") {
		float fac0, fac1, fac2, fac3;
		float a;

		color col = color colorband(coba, tin, a);
		fac0 = comp(col, 0) + comp(col, 1) + comp(col, 2);
		col = color colorband(coba, xcomp(N), a);
		fac1 = comp(col, 0) + comp(col, 1) + comp(col, 2);
		col = color colorband(coba, ycomp(N), a);
		fac2 = comp(col, 0) + comp(col, 1) + comp(col, 2);
		col = color colorband(coba, zcomp(N), a);
		fac3 = comp(col, 0) + comp(col, 1) + comp(col, 2);

		n = normal(0.3333*(fac0 - fac1), 0.3333*(fac0 - fac2), 0.3333*(fac0 - fac3));
	} else {
	  n = normal(tin, tin, tin) - n;
	}

	return n;
}

float bricont(float tin, bright, contr)
{
	return clamp((tin - 0.5) * contr + bright - 0.5, 0, 1);
}

color bricontrgb(color rgb; float bright, contr; color fac)
{
	/* Blender also bothers to place a lower clamp at 0 */
	return fac * ((rgb - 0.5)*contr + bright - 0.5);
}

normal tintonor(float tin; normal Nf)
{
  float co= 0.5*cos(tin-0.5);
  float si= 0.5*sin(tin-0.5);
  float f1, f2, f3;

  /*  normal nor;

  f1 = xcomp(Nf); f2 = ycomp(Nf);
  setxcomp(nor, f1*co+f2*si);
  setycomp(nor, f2*co-f1*si);

  f1 = ycomp(Nf); f2 = xcomp(Nf);
  setycomp(nor, f1*co+f2*si);
  setzcomp(nor, f2*co-f1*si);*/

  f1 = xcomp(Nf);
  f2 = ycomp(Nf);
  f3 = zcomp(Nf);

  normal rN = normal(f1 * co + f2 * si,
		     f2 * co + f3 * si,
		     f3 * co - f2 * si);

	return rN;
}

float Blnd_TexCloudsTin(point P;
						float isHard, noiseSize, noiseDepth;
						string noiseType)
{
	return blenturb(noiseSize, P, noiseDepth, isHard, Blnd_NoiseType(noiseType));
}

normal Blnd_TexCloudsNor(float tin;
			point P;
		     float isHard, noiseSize, noiseDepth;
		     string noiseType; float nabla;
		     string coba)
{
  normal N = normal(blenturb(noiseSize, P + vector(nabla, 0, 0), noiseDepth,
			      isHard, Blnd_NoiseType(noiseType)),
		    blenturb(noiseSize, P + vector(0, nabla, 0), noiseDepth,
			      isHard, Blnd_NoiseType(noiseType)),
		    blenturb(noiseSize, P + vector(0, 0, nabla), noiseDepth,
			      isHard, Blnd_NoiseType(noiseType)) );

  return norm_derivate(tin, N, coba);
}

color Blnd_TexCloudsRGB(float tin; output float a;
						point P;
						float isHard, noiseSize, noiseDepth;
						string noiseType)
{
	a = 1;
  point P2 = point(ycomp(P), xcomp(P), zcomp(P));
  point P3 = point(ycomp(P), zcomp(P), xcomp(P));
  return color(
	       tin,
	       blenturb(noiseSize, P2, noiseDepth, isHard, Blnd_NoiseType(noiseType)),
	       blenturb(noiseSize, P3, noiseDepth, isHard, Blnd_NoiseType(noiseType)));
}

float Blnd_TexWoodTin(
					point P;
					string mode, wavetype;
					float isHard, noiseSize, turbulence;
					string noiseType)
{
	float tin = 0;

	if (mode == "bands")
		tin = Blnd_Waveform(wavetype, (xcomp(P)+ycomp(P)+zcomp(P)) * 10);
	else if (mode == "rings")
		tin = Blnd_Waveform(wavetype, length(vector P) * 20);
	else if (mode == "bandnoise")
		tin = Blnd_Waveform(wavetype, (xcomp(P)+ycomp(P)+zcomp(P)) * 10 +
			turbulence * blennoise(noiseSize, P, isHard,
			Blnd_NoiseType(noiseType)));
	else if (mode == "ringnoise")
		tin = Blnd_Waveform(wavetype, length(vector P) * 20 + turbulence *
			blennoise(noiseSize, P, isHard, Blnd_NoiseType(noiseType)));

	return tin;
}

normal Blnd_TexWoodNor(float tin;
			point P;
		   string mode, wavetype;
		   float isHard, noiseSize, turbulence;
		   string noiseType; float nabla;
		   string coba)
{
  normal N = normal(Blnd_TexWoodTin(P + vector(nabla,0,0),mode,wavetype,
									isHard,noiseSize,turbulence,noiseType),
					Blnd_TexWoodTin(P + vector(0,nabla,0),mode,wavetype,
									isHard,noiseSize,turbulence,noiseType),
					Blnd_TexWoodTin(P + vector(0,0,nabla),mode,wavetype,
									isHard,noiseSize,turbulence,noiseType));

  return norm_derivate(tin, N, coba);
}

normal Blnd_TexStucciNor(output float tin;
			point P;
			 float stype, isHard, noiseSize, turbulence;
			 string noiseType;
			 string coba)
{
	tin = blennoise(noiseSize, P, isHard, Blnd_NoiseType(noiseType));
  float ofs = turbulence / 200.0;

  if (stype > 0)
    ofs *= tin*tin;

  normal N = normal(blennoise(noiseSize, P + vector(ofs, 0, 0), isHard,
			      Blnd_NoiseType(noiseType)),
		    blennoise(noiseSize, P + vector(0, ofs, 0), isHard,
			      Blnd_NoiseType(noiseType)),
		    blennoise(noiseSize, P + vector(0, 0, ofs), isHard,
			      Blnd_NoiseType(noiseType)) );

  N = norm_derivate(tin, N, coba);

  if (stype == 2)
    N = -N;

  return N;
}

color Blnd_TexMagicRGB(output float tin, a; point P; float n, turbulence)
{
  float x, y, z;
  float turb = turbulence / 5.0;

  x = sin( (xcomp(P)+ycomp(P)+zcomp(P)) * 5);
  y = cos( (-xcomp(P)+ycomp(P)-zcomp(P)) * 5);
  z = -cos( (-xcomp(P)-ycomp(P)+zcomp(P)) * 5);

  if(n>0) {
    x*= turb;
    y*= turb;
    z*= turb;
    y= -cos(x-y+z);
    y*= turb;
    if(n>1) {
      x= cos(x-y-z);
      x*= turb;
      if(n>2) {
	z= sin(-x-y-z);
	z*= turb;
	if(n>3) {
	  x= -cos(-x+y-z);
	  x*= turb;
	  if(n>4) {
	    y= -sin(-x+y+z);
	    y*= turb;
	    if(n>5) {
	      y= -cos(-x+y+z);
	      y*= turb;
	      if(n>6) {
		x= cos(x+y+z);
		x*= turb;
		if(n>7) {
		  z= sin(x+y-z);
		  z*= turb;
		  if(n>8) {
		    x= -cos(-x-y+z);
		    x*= turb;
		    if(n>9) {
		      y= -sin(x-y+z);
		      y*= turb;
		    }
		  }
		}
	      }
	    }
	  }
	}
      }
    }
  }

  if (turb != 0) {
    turb *= 2;
    x /= turb;
    y /= turb;
    z /= turb;
  }

	color rgb = color(0.5 - x, 0.5 - y, 0.5 - z);

	tin = 0.3 * (comp(rgb, 0) + comp(rgb, 1) + comp(rgb, 2));

	a = 1;

  return rgb;
}

float Blnd_TexMarbleTin(point P;
		     float sharpness, isHard;
		     string waveType;
		     float noiseSize, noiseDepth, turbulence;
		     string noiseType)
{
  float n = 5 * (xcomp(P) + ycomp(P) + zcomp(P));
  float mi = n + turbulence *
    blenturb(noiseSize, P, noiseDepth, isHard, Blnd_NoiseType(noiseType));
  mi = Blnd_Waveform(waveType, mi);
  if (sharpness == 1)
    mi = sqrt(mi);
  else if (sharpness == 2)
    mi = sqrt(sqrt(mi));
  return mi;
}

normal Blnd_TexMarbleNor(float tin; point P;
			 float sharpness, isHard;
			 string waveType;
			 float noiseSize, noiseDepth, turbulence;
			 string noiseType; float nabla;
			 string coba)
{
  normal N = normal(Blnd_TexMarbleTin(P + vector(nabla, 0, 0), sharpness,
				   isHard, waveType, noiseSize, noiseDepth,
				   turbulence, noiseType),
		    Blnd_TexMarbleTin(P + vector(0, nabla, 0), sharpness,
				   isHard, waveType, noiseSize, noiseDepth,
				   turbulence, noiseType),
		    Blnd_TexMarbleTin(P + vector(0, 0, nabla), sharpness,
				   isHard, waveType, noiseSize, noiseDepth,
				   turbulence, noiseType));

  return norm_derivate(tin, N, coba);
}

float Blnd_TexBlendTin(
		    point P;
		    string mode;
		    float flipXY;
		    )
{
  float f = 0;
  float x, y;

  if (flipXY == 0)
    {
      x = xcomp(P);
      y = ycomp(P);
    } else {
      x = ycomp(P);
      y = xcomp(P);
    }

  if (mode == "lin")
    f = (1 + x) / 2;
  else if (mode == "quad")
    {
      f = (1 + x) / 2;
      if (f < 0) f = 0;
      else f *= f;
    }
  else if (mode == "ease")
    {
      f = (1 + x) / 2;
      if (f < 0) f = 0;
      else if (f > 1) f = 1;
      else {
	float t=f*f;
	f=(3*t-2*t*f);
      }
    }
  else if (mode == "diag")
    {
      f = (2+x+y)/4;
    }
  else if (mode == "sphere")
    {
      f = 1 - length(vector P);
      if (f < 0) f = 0;
    }
  else if (mode == "halo")
    {
      f = 1 - length(vector P);
      if (f < 0) f = 0;
      f *= f;
    }

  f = clamp(f, 0, 1);

  return f;
}

color Blnd_TexImageRGB(output float a; point P; string filename)
{
	a = texture(filename[3], xcomp(P), ycomp(P), "fill", 1);
	return texture(filename, xcomp(P), ycomp(P));
}
