# Blender material->Renderman shader exporter
# Version 0.3

import Blender
from Blender import Material, Texture, Scene

obj = Scene.getCurrent().getActiveObject()
try:
	mat = obj.getMaterials()[0]
except:
	mat = obj.getData().getMaterials()[0]
outPath = '/home/sapir/Renderman/Blender/material.sl'

woodModeDict = { 0:'"bands"', 1:'"rings"', 2:'"bandnoise"', 3:'"ringnoise"' }
blendModeDict = { 0:'"lin"', 1:'"quad"', 2:'"ease"',
    3:'"diag"', 4:'"sphere"', 5:'"halo"' }
hardDict = { 'hard':'1', 'soft':'0' }
noiseTypeDict = {
    0:'"Default"',
    1:'"OriginalPerlin"',
    2:'"ImprovedPerlin"',
    3:'"VoronoiF1"',
    4:'"VoronoiF2"',
    5:'"VoronoiF3"',
    6:'"VoronoiF4"',
    7:'"VoronoiF2-F1"',
    8:'"VoronoiCrackle"',
    14:'"CellNoise"' }
waveTypeDict = { 0:'"sin"', 1:'"saw"', 2:'"tri"' }
axis = [ 'x', 'y', 'z' ]
zeroLimit = 0.001

def strf(x):
    return '%.3f' % x

def writeln(f,c=None):
    if c != None:
        f.write(c)
    f.write('\n')


# TODO: blending functions would benefit from not calculating things twice
	# (in Blender things are stored in variables and remembered,
	# here I keep substituting expressions over and over so they
	# most likely get recalculated)

def rgbBlend(f, inp, tex, out, fact, facg, blendmode):
	# The function currently assumes inp=out, particularly for lighten/darken. :/
	if blendmode == Texture.BlendMode['MIX']:
		writeln(f,'\t%s = mix(%s, %s, %s * %s);' % (out, inp, tex, fact, facg))
	elif blendmode == Texture.BlendMode['MULTIPLY']:
		writeln(f,'\t%s = (1 - %s + %s * %s * %s) * %s;' %
			(out, facg, fact, facg, tex, inp))
	elif blendmode == Texture.BlendMode['SCREEN']:
		writeln(f,'\t%s = 1 - (1-%s + %s*%s*(1-%s))*(1-%s);\n' % (out, facg,
			fact, facg, tex, inp))
	elif blendmode == Texture.BlendMode['ADD']:
		writeln(f,'\t%s = %s + %s * %s * %s;' % (out, inp, tex, fact, facg))
	elif blendmode == Texture.BlendMode['SUBTRACT']:
		writeln(f,'\t%s = %s - %s * %s * %s;' % (out, inp, tex, fact, facg))
	elif blendmode == Texture.BlendMode['DIVIDE']:
		for i in range(3):
			texi = 'comp(%s, %d)' % (tex, i)
			outi = 'comp(%s, %d)' % (out, i)
			inpi = 'comp(%s, %d)' % (inp, i)
			writeln(f, '\tif (%s != 0) {' % texi)
			writeln(f, '\t\tsetcomp(%s, %d,' % (out, i))
			writeln(f, '\t\t\tmix(%s, %s / %s, %s * %s));' %
				(inpi, inpi, texi, fact, facg))
			writeln(f, '\t}')
	elif blendmode == Texture.BlendMode['DIFFERENCE']:
		for i in range(3):
			texi = 'comp(%s, %d)' % (tex, i)
			outi = 'comp(%s, %d)' % (out, i)
			inpi = 'comp(%s, %d)' % (inp, i)
			writeln(f, '\tsetcomp(%s, %d,' % (out, i))
			writeln(f, '\t\tmix(%s, abs(%s - %s),' % (inpi, texi, inpi))
			writeln(f, '\t\t%s*%s));' % (fact, facg))
	elif blendmode == Texture.BlendMode['LIGHTEN']:
		for i in range(3):
			texi = 'comp(%s, %d)' % (tex, i)
			inpi = 'comp(%s, %d)' % (inp, i)
			writeln(f, '\tif (%s*%s*%s > %s)' % (fact,facg,texi,inpi))
			writeln(f, '\t{ setcomp(%s, %s, %s); }' % (out, i, texi))
	elif blendmode == Texture.BlendMode['DARKEN']:
		for i in range(3):
			texi = 'comp(%s, %d)' % (tex, i)
			inpi = 'comp(%s, %d)' % (inp, i)
			writeln(f, '\tif (%s*%s*%s < %s)' % (fact,facg,texi,inpi))
			writeln(f, '\t{ setcomp(%s, %s, %s); }' % (out, i, texi))

def valBlend(f, inp, tex, out, fact, facg, blendmode, flip):
	# The function currently assumes inp=out, particularly for lighten/darken. :/
	if flip > 0:
		_fact = '(%s * %s)' % (fact, facg)
		_facm = '(1 - (%s * %s))' % (fact, facg)
	else:
		_facm = '(%s * %s)' % (fact, facg)
		_fact = '(1 - (%s * %s))' % (fact, facg)

	if blendmode == Texture.BlendMode['MIX']:
		# Do this separately because Renderman has a shadeop so
			# flipping is a special case
		if flip > 0:
			writeln(f,'\t%s = mix(%s, %s, %s * %s);' % (out, inp, tex, fact, facg))
		else:
			writeln(f,'\t%s = mix(%s, %s, %s * %s);' % (out, tex, inp, fact, facg))
	elif blendmode == Texture.BlendMode['MULTIPLY']:
		_facm = '(1 - (%s))' % facg
		writeln(f,'\t%s = (%s+%s*%s)*%s;' % (out, _facm, _fact, tex, inp))
	elif blendmode == Texture.BlendMode['SCREEN']:
		_facm = '(1 - (%s))' % facg
		writeln(f,'\t%s=1-(%s+%s*(1-%s))*(1-%s);' %
			(out, _facm, _fact, tex, inp))
	elif blendmode == Texture.BlendMode['ADD']:
		writeln(f,'\t%s = %s*%s + %s;' % (out, _fact, tex, inp))
	elif blendmode == Texture.BlendMode['SUBTRACT']:
		writeln(f,'\t%s = %s*%s - %s;' % (out, _fact, tex, inp))
	elif blendmode == Texture.BlendMode['DIVIDE']:
		# Same idea as mix
		writeln(f, '\tif (%s != 0)' % tex)
		if flip > 0:
			writeln(f,'\t{ %s = mix(%s, %s/%s, %s*%s); }' %
				(out, inp, inp, tex, fact, facg))
		else:
			writeln(f,'\t{ %s = mix(%s/%s, %s, %s*%s); }' %
				(out, inp, tex, tex, fact, facg))
	elif blendmode == Texture.BlendMode['DIFFERENCE']:
		# Same idea as mix
		if flip > 0:
			writeln(f,'\t%s = mix(%s, abs(%s - %s), %s * %s);' %
				(out, inp, tex, inp, fact, facg))
		else:
			writeln(f,'\t%s = mix(abs(%s - %s), %s, %s * %s);' %
				(out, tex, inp, inp, fact, facg))
	elif blendmode == Texture.BlendMode['LIGHTEN']:
		writeln(f,'\tif (%s*%s > %s) { %s = %s * %s; }' %
			(_fact, tex, inp, out, _fact, tex))
	elif blendmode == Texture.BlendMode['DARKEN']:
		writeln(f,'\tif (%s*%s < %s) { %s = %s * %s; }' %
			(_fact, tex, inp, out, _fact, tex))

def getCoba(coba):
	cobaStr = '"L '
	for i in range(len(coba)):
		c = coba[i]
		cobaStr += (' %.3f %.3f %.3f %.3f %.3f' %
			(c[0][0], c[0][1], c[0][2], c[0][3], c[1]))
	cobaStr += '"'
	return cobaStr

f = open(outPath, "w")

writeln(f,'/* textures and local illumination models */')
writeln(f,'#include "locillum.h"')
writeln(f,'#include "btexs2.h"')
writeln(f,'#include "bli.h"')
writeln(f,'#include "bimage.h"')

#matName = mat.getName()
matName = 'Material'
if False:#mat.vcolLight:
	paramList = 'varying color vclight = color(1,1,1);'
else:
	paramList = ''
writeln(f,'surface %s(%s)' % ('Material', paramList))
writeln(f,'{')

#writeln(f, '\tcolor Col = Cs;')
writeln(f, '\tcolor Col = color(%.3f, %.3f, %.3f);' % tuple(mat.rgbCol))
writeln(f, '\tcolor Csp = color(%.3f, %.3f, %.3f);' % tuple(mat.specCol))
writeln(f, '\tcolor Cmir = color(%.3f, %.3f, %.3f);' % tuple(mat.mirCol))
writeln(f, '\tfloat hard = %.3f;' % mat.hard)
writeln(f, '\tfloat Ka = %.3f, Kd = %.3f, Ks = %.3f, Kt = %.3f;' %
	(mat.amb, mat.ref, mat.spec, mat.translucency))

# For Aqsis:
# writeln(f, '\tnormal Nf = normalize(ntransform("camera", N));')
# For 3delight:
# writeln(f, '\tnormal Nf = -normalize(ntransform("camera", N));')
writeln(f, '\tnormal Nf = N;')
writeln(f, '\tsetzcomp(Nf, -zcomp(Nf));')
writeln(f, '\tNf = normalize(ntransform("camera", faceforward(Nf, I)));')
writeln(f)
writeln(f, '\tOi = Os;')
writeln(f)


# Textures

writeln(f, '\tpoint Pcur;')
writeln(f, '\tfloat tin;')
writeln(f, '\tcolor rgb;')
writeln(f, '\tfloat a;')
writeln(f, '\tnormal nor;')
writeln(f, '\tfloat Tnor = 1;')
writeln(f, '\tfloat stencilTin = 1;')
writeln(f, '\tfloat tmp;')
writeln(f, '\tvector warpVec;')
writeln(f, '\tstring coba = "";')
writeln(f)

writeln(f, 'float debug = 0;')
writeln(f, 'point S = transform("raster",P);')
writeln(f, 'if (abs(151-xcomp(S))<0.5 && abs(123-ycomp(S))<0.5)')
writeln(f, '{ debug = 1; }')

warpDone = False
stencilDone = False

mtexs = mat.getTextures()
for mtex in mtexs:
	if mtex == None:
		continue
	tex = mtex.tex
	texType = tex.getType()

	writeln(f, '/* %s of type %s */' % (tex.name, texType))

	if mtex.texco == Texture.TexCo['ORCO']:
		# We *could* use object here (and everywhere else!!),
			# but shader is nicer :)
		writeln(f,'\tPcur = transform("shader", P);')
	elif mtex.texco == Texture.TexCo['GLOB']:		
		writeln(f,'\tPcur = transform("world", P);')
#		writeln(f,'\tsetxcomp(Pcur, -xcomp(Pcur));')
	elif mtex.texco == Texture.TexCo['WIN']:
		writeln(f,'\tPcur = transform("NDC", P);')
		writeln(f,'\tsetxcomp(Pcur, (xcomp(Pcur) * 2) - 1);')
		writeln(f,'\tsetycomp(Pcur, (ycomp(Pcur) * -2) + 1);')
		writeln(f,'\tsetzcomp(Pcur, 0);')

	elif mtex.texco == Texture.TexCo['NOR']:
		writeln(f,'\tPcur = point Nf;')
		writeln(f,'\tif (debug != 0) { printf("  Nor Pcur %p\\n", Pcur); }')
		# Pcur = -Pcur
		#setzcomp(Pcur, -zcomp(Pcur))
#		writeln(f,'\tsetxcomp(Pcur,-xcomp(Pcur));')
		#writeln(f,'\tsetycomp(Pcur,-ycomp(Pcur));')

	# TODO: the following don't work
	elif mtex.texco == Texture.TexCo['UV']:
		writeln(f,'\tPcur = point(s, t, 0);')
	elif mtex.texco == Texture.TexCo['REFL']:
		# Use nor as a temporary variable
#		writeln(f,'\tnor = normalize(ntransform("camera", "current", Nf));')
		# NOTE: copy of code at shading section!
		writeln(f,'\tnor = ntransform("camera", "current", Nf);')
		writeln(f,'\tsetzcomp(nor, -zcomp(nor));')
		writeln(f,'\tnor = normalize(nor);')
		writeln(f,'\tPcur = point normalize(reflect(I, nor));')
		writeln(f,'\tsetzcomp(Pcur,-zcomp(Pcur));')
	elif mtex.texco == Texture.TexCo['OBJECT']:
		pass
	elif mtex.texco == Texture.TexCo['STICKY']:
		pass

	if warpDone:
		writeln(f, '\tPcur += warpVec;')

	# TODO: images have special projections

	# NOTE: I'm assuming projections x,y,z are 1,2,3
		# (so does the Blender code actually)
	if not (mtex.xproj == Texture.Proj['X'] and
			mtex.yproj == Texture.Proj['Y'] and
			mtex.zproj == Texture.Proj['Z']):
		proj = 'point('
		if mtex.xproj == Texture.Proj['NONE']:
			proj = proj + '0, '
		else:
			proj = proj + ('%scomp(Pcur), ' % axis[mtex.xproj - 1])
		if mtex.yproj == Texture.Proj['NONE']:
			proj = proj + '0, '
		else:
			proj = proj + ('%scomp(Pcur), ' % axis[mtex.yproj - 1])
		if mtex.zproj == Texture.Proj['NONE']:
			proj = proj + '0)'
		else:
			proj = proj + ('%scomp(Pcur))' % axis[mtex.zproj - 1])
		writeln(f, '\tPcur = %s;' % proj)

	for i in range(3):
		dx = mtex.ofs[i]
		sx = mtex.size[i]
		ax = axis[i]
		hasOfs = (abs(dx) >= zeroLimit)
		hasSize = (abs(sx - 1) >= zeroLimit)
		if hasOfs and hasSize:
			writeln(f,'\tset%scomp(Pcur, (%scomp(Pcur) + %.3f) * %.3f);\n'
				% (ax, ax, dx, sx))
		elif hasSize:
			writeln(f,'\tset%scomp(Pcur, %scomp(Pcur) * %.3f);\n' % (ax, ax, sx))
		elif hasOfs:
			writeln(f,'\tset%scomp(Pcur, %scomp(Pcur) + %.3f);\n' % (ax, ax, dx))


	wantNor = (mtex.mtNor != 0 or mtex.mtDisp != 0 or mtex.mtWarp != 0)
#	hasTin = False
	hasRGB = False
	hasNor = False
	hasAlpha = False
	# TODO: fix next line
	texHasColorBand	= (tex.colorband != None) # ?

	if texHasColorBand:
		writeln(f, '\tcoba = %s;' % getCoba(tex.colorband))

	if texType == 'Wood':
		argList = [ 'Pcur', woodModeDict[tex.stype], waveTypeDict[tex.wavetype],
			hardDict[tex.noiseType], strf(tex.noiseSize), strf(tex.turbulence),
			noiseTypeDict[tex.noiseBasis] ]
		writeln(f, '\ttin = Blnd_TexWoodTin(%s);' % ','.join(argList))
		if wantNor:
			hasNor = True
			argList.insert(0, 'tin')
			argList.append(str(0.025))
			argList.append('coba')
			writeln(f, '\tnor = Blnd_TexWoodNor(%s);' % ','.join(argList))
#		else:
#			hasTin = True
		writeln(f, '\ttin = bricont(tin, %.3f, %.3f);' %
			(tex.brightness, tex.contrast))
	elif texType == 'Stucci':
		if wantNor:
			hasNor = True
		# wantNor for nor, texHasColorband for tin.
		if wantNor or texHasColorBand:
			argList = [ 'tin', 'Pcur', str(tex.stype), hardDict[tex.noiseType],
				strf(tex.noiseSize), strf(tex.turbulence),
				noiseTypeDict[tex.noiseBasis], 'coba' ]
			writeln(f, '\tnor = Blnd_TexStucciNor(%s);' % ','.join(argList))
		else:
			writeln(f, '\ttin = 0;')
	elif texType == 'Clouds':
		argList = [ 'Pcur', hardDict[tex.noiseType], strf(tex.noiseSize),
			strf(tex.noiseDepth), noiseTypeDict[tex.noiseBasis] ]
		writeln(f, '\ttin = Blnd_TexCloudsTin(%s);' % ','.join(argList))
		if wantNor:
			argList = [ 'tin', 'Pcur', hardDict[tex.noiseType],
				strf(tex.noiseSize), strf(tex.noiseDepth),
				noiseTypeDict[tex.noiseBasis], strf(0.025), 'coba' ]
			writeln(f, '\tnor = Blnd_TexCloudsNor(%s);' % ','.join(argList))
			hasNor = True
		if tex.stype == 1:
			argList = [ 'tin', 'a', 'Pcur', hardDict[tex.noiseType],
				strf(tex.noiseSize), strf(tex.noiseDepth),
				noiseTypeDict[tex.noiseBasis] ]
			writeln(f, '\trgb = Blnd_TexCloudsRGB(%s);' % ','.join(argList))
			writeln(f, '\trgb = bricontrgb(rgb, %.3f, %.3f, color(%.3f, %.3f, %.3f));'
				% (tex.brightness, tex.contrast,
				tex.rgbCol[0], tex.rgbCol[1], tex.rgbCol[2]))
			hasRGB = True
		writeln(f, '\ttin = bricont(tin, %.3f, %.3f);' %
			(tex.brightness, tex.contrast))
	elif texType == 'Magic':
		argList = [ 'tin', 'a', 'Pcur', strf(tex.noiseDepth), strf(tex.turbulence) ]
		writeln(f, '\trgb = Blnd_TexMagicRGB(%s);' % ','.join(argList))
		writeln(f, '\trgb = bricontrgb(rgb, %.3f, %.3f, color(%.3f, %.3f, %.3f));'
			% (tex.brightness, tex.contrast,
			tex.rgbCol[0], tex.rgbCol[1], tex.rgbCol[2]))
		hasRGB = True
	elif texType == 'None':
		writeln(f, '\ttin = 0;')
	elif texType == 'Marble':
		argList = [ 'Pcur', str(tex.stype), hardDict[tex.noiseType],
			waveTypeDict[tex.wavetype], strf(tex.noiseSize),
			strf(tex.noiseDepth), strf(tex.turbulence),
			noiseTypeDict[tex.noiseBasis] ]
		writeln(f, '\ttin = Blnd_TexMarbleTin(%s);' % ','.join(argList))
		if wantNor:
			argList.insert(0, 'tin')
			argList.append(str(0.025))
			argList.append('coba')			
			writeln(f, '\tnor = Blnd_TexMarbleNor(%s);' % ','.join(argList))
			hasNor = True
		writeln(f, '\ttin = bricont(tin, %.3f, %.3f);' %
			(tex.brightness, tex.contrast))
	elif texType == 'Blend':
		# TODO: flipxy
		argList = [ 'Pcur', blendModeDict[tex.stype], str(0) ]
		writeln(f, '\ttin = Blnd_TexBlendTin(%s);' % ','.join(argList))
		writeln(f, '\ttin = bricont(tin, %.3f, %.3f);' %
			(tex.brightness, tex.contrast))
	elif texType == 'Image':
		imageFN = tex.name + '.tiff' #tex.image.filename
		# TODO: use Texture.Mapping dictionary
		if mtex.mapping == 0:
			writeln(f, '\tPcur = (Pcur + vector(1,1,0)) / 2;')
		elif mtex.mapping == 1:
			# TODO: object cubemap, global cubemap
			writeln(f, '\tPcur = Blnd_CubeMap(Pcur);')
		elif mtex.mapping == 2:
			writeln(f, '\tPcur = Blnd_TubeMap(Pcur);')
		elif mtex.mapping == 3:
			writeln(f, '\tPcur = Blnd_SphereMap(Pcur);')
		writeln(f, '\trgb = texture("%s", xcomp(Pcur), 1-ycomp(Pcur));' % imageFN)
		if tex.imageFlags & Texture.ImageFlags['USEALPHA']:
			writeln(f, '\ta = texture("%s"[3], xcomp(Pcur), 1-ycomp(Pcur), "fill", 1);'
				% imageFN)
			hasAlpha = True
		elif tex.imageFlags & Texture.ImageFlags['CALCALPHA']:
			writeln(f, '\ta = max(comp(rgb, 0), comp(rgb, 1), comp(rgb, 2));')
		writeln(f, '\ttin = a;')
		if tex.flags & Texture.Flags['NEGALPHA']:
			writeln(f, '\ta = 1 - a;')
		writeln(f, '\trgb = bricontrgb(rgb, %.3f, %.3f, color(%.3f, %.3f, %.3f));'
			% (tex.brightness, tex.contrast,
			tex.rgbCol[0], tex.rgbCol[1], tex.rgbCol[2]))
		hasRGB = True

	if texHasColorBand:
		writeln(f, '\trgb = color colorband(coba, tin, a);')
		hasRGB = True
		hasAlpha = True

	if hasRGB and mtex.noRGB:
		writeln(f, '\ttin = 0.35*comp(rgb,0)+0.45*comp(rgb,1)+0.2*comp(rgb,2);')
		hasRGB = False

	if mtex.neg:
		if hasRGB:
			writeln(f, '\trgb = color(1) - rgb;')
		writeln(f, '\ttin =  1 - tin;')

	if mtex.stencil:
		if hasRGB:
			writeln(f, '\ttmp = a;')
			writeln(f, '\ta *= stencilTin;')
			writeln(f, '\tstencilTin *= tmp;')
		else:
			writeln(f, '\ttmp = tin;')
			writeln(f, '\ttin *= stencilTin;')
			writeln(f, '\tstencilTin *= tmp;')
		stencilDone = True
	else:
		if stencilDone:
			writeln(f, '\ta *= stencilTin;')
			writeln(f, '\tTnor *= stencilTin;')
			writeln(f, '\ttin *= stencilTin;')

	if wantNor:
		if not hasNor:
			if hasRGB:
				writeln(f, '\tnor = normal rgb;')
			else:
				writeln(f, '\tnor = tintonor(tin, Nf);')

		if mtex.mtWarp == 1:
			if mtex.warpfac >= zeroLimit:
				# NOTE: no ntransform() for nor for warpVec, because
					# we want the pure result of the texture's Nor function,
					# not something we can calculate lighting with.
				writeln(f, '\twarpVec = %.3f * nor;' % mtex.warpfac)
				warpDone = True
			else:
				warpDone = False

		if mtex.correctNor:
			if (mtex.texco == Texture.TexCo['ORCO'] or
				mtex.texco == Texture.TexCo['UV']):

				writeln(f, '\tnor = ntransform("world", "camera", nor);')


	if mtex.mtCol == 1 or mtex.mtCsp == 1 or mtex.mtCmir == 1:
		if hasRGB:
			blendColor = 'rgb'
			if mtex.mtAlpha != 0:
				writeln(f, '\ttin = stencilTin;')
			else:
				writeln(f, '\ttin = a;')
		else:
			blendColor = 'color(%.3f, %.3f, %.3f)' % mtex.col
		if mtex.mtCol == 1:
			rgbBlend(f, 'Col', blendColor, 'Col', 'tin', mtex.colfac,
				mtex.blendmode)
		if mtex.mtCsp == 1:
			rgbBlend(f, 'Csp', blendColor, 'Csp', 'tin', mtex.colfac,
				mtex.blendmode)
		if mtex.mtCmir == 1:
			if texType == 'EnvMap' and mtex.blendmode == Texture.BlendMode['MIX']:
				# TODO: do stuff!
				pass
			else:
				rgbBlend(f, 'Cmir', blendColor, 'Cmir', 'tin', mtex.colfac,
					mtex.blendmode)

	if mtex.mtNor != 0 and wantNor:
		# Nor almost works. Blender output seems blurry.
		# NOTE: Blender flips the normal factor, but for some reason
			# we shouldn't
		# NOTE: yafray undoes it for stucci - ??
		norfac = mtex.norfac * -mtex.mtNor
		# TODO: image normal map
		writeln(f, '\tNf += Tnor*%s*nor;' % norfac)
		writeln(f, '\tNf = normalize(Nf);')
		# NOTE: Blender recalculates reflection cache vector here
			# we don't need to do this because we're not caching
			# the reflection vector

	if mtex.mtDisp != 0:
		# TODO: displacement
		pass

	# We're up to the variables now.
	if (mtex.mtRef != 0 or mtex.mtSpec != 0 or mtex.mtEmit != 0 or
		mtex.mtAlpha != 0 or mtex.mtHard != 0  or mtex.mtTranslu != 0 or
		mtex.mtAmb != 0):
		
		if hasRGB:
			if hasAlpha:
				writeln(f, '\ttin = a;\n')
			else:
				writeln(f, '\ttin = 0.35*comp(rgb,0)+0.45*comp(rgb,1)+0.2*comp(rgb,2);')

		if mtex.mtRef != 0:
			valBlend(f, 'Kd', mtex.dvar, 'Kd', 'tin', mtex.varfac,
				mtex.blendmode, mtex.mtRef)
			writeln(f, '\tif (Kd < 0) { Kd = 0; }')
		if mtex.mtSpec != 0:
			valBlend(f, 'Ks', mtex.dvar, 'Ks', 'tin', mtex.varfac,
				mtex.blendmode, mtex.mtSpec)
			writeln(f, '\tif (Ks < 0) { Ks = 0; }')
		if mtex.mtEmit != 0:
			valBlend(f, 'emit', mtex.dvar, 'emit', 'tin', mtex.varfac,
				mtex.blendmode, mtex.mtEmit)
			writeln(f, '\tif (emit < 0) { emit = 0; }')
		if mtex.mtAlpha != 0:
			# Blender uses a single value but Renderman uses 3!
			# Let's do it the Renderman way...
			blendColor = ('color(%.3f, %.3f, %.3f)' %
				(mtex.dvar, mtex.dvar, mtex.dvar))
#			rgbBlendFlip(f, 'Oi', blendColor, 'Oi', 'tin', mtex.varfac,
				#mtex.blendmode, mtex.mtAlpha)
			rgbBlend(f, 'Oi', blendColor, 'Oi', 'tin', mtex.varfac,
				mtex.blendmode)
			# Blender places a lower clamp of 0 on alpha.
			# I don't think Renderman minds.
		if mtex.mtHard != 0:
			writeln(f, '\thard /= 128;')
			valBlend(f, 'hard', mtex.dvar, 'hard', 'tin', mtex.varfac,
				mtex.blendmode, mtex.mtHard)
			writeln(f, '\thard *= 128;')
			writeln(f, '\thard = clamp(hard, 1, 511);')
		if mtex.mtTranslu != 0:
			valBlend(f, 'Kt', mtex.dvar, 'Kt', 'tin', mtex.varfac,
				mtex.blendmode, mtex.mtTranslu)
			writeln(f, '\tKt = clamp(Kt, 0, 1);')
		if mtex.mtAmb != 0:
			valBlend(f, 'Ka', mtex.dvar, 'Ka', 'tin', mtex.varfac,
				mtex.blendmode, mtex.mtAmb)
			writeln(f, '\tKa = clamp(Ka, 0, 1);')

	writeln(f)



# NOTE: if mat is shadeless or vclight, there's no need to calculate hard etc.
if False:#mat.shadeless:
	writeln(f,'\tCi = Col;')
	writeln(f,'\tCi *= Oi;')
elif False:#mat.vcolLight:
	writeln(f,'\tCi = vclight * Col;')
	writeln(f,'\tCi *= Oi;')
else:
	if mat.specTransp >= zeroLimit:
		writeln(f,'\ttmp = max(comp(Csp,0), comp(Csp,1), comp(Csp,2));')
		if abs(mat.specTransp - 1) >= zeroLimit:
			writeln(f,'\ttmp *= %.3f' % mat.specTransp)
		writeln(f, '\tif (tmp > 1) { tmp = 1; }')
		writeln(f, '\tOi = mix(Oi, color(1), tmp);')

	writeln(f, '\tNf = ntransform("camera", "current", Nf);')
	writeln(f, '\tsetzcomp(Nf, -zcomp(Nf));')
	writeln(f, '\tNf = normalize(Nf);')

	diffShader = mat.diffuseShader
	if diffShader == Material.Shaders['DIFFUSE_LAMBERT']:
		diffuseFunc = 'diffuse(Nf)'
		transluFunc = 'diffuse(-Nf)'
	elif diffShader == Material.Shaders['DIFFUSE_ORENNAYAR']:
		diffuseFunc = ('LocIllumOrenNayar(Nf, -normalize(I), %.3f)' %
			mat.roughness)
		transluFunc = ('LocIllumOrenNayar(-Nf, -normalize(I), %.3f)' % 
			mat.roughness)
	elif diffShader == Material.Shaders['DIFFUSE_TOON']:
		diffuseFunc = ('Blnd_LIToonDiffuse(Nf, -normalize(I), %.3f, %.3f)' %
			(mat.diffuseSize, mat.diffuseSmooth))
		transluFunc = ('Blnd_LIToonDiffuse(-Nf, -normalize(I), %.3f, %.3f)' %
			(mat.diffuseSize, mat.diffuseSmooth))
	elif diffShader == Material.Shaders['DIFFUSE_MINNAERT']:
		diffuseFunc = ('Blnd_LIMinnaert(Nf, -normalize(I), %.3f)' %
			mat.diffuseDarkness)
		transluFunc = ('Blnd_LIMinnaert(-Nf, -normalize(I), %.3f)' %
			mat.diffuseDarkness)

	specShader = mat.specShader
	if specShader == Material.Shaders['SPEC_COOKTORR']:
		specFunc = "Blnd_LICookTorr(Nf, -normalize(I), hard)"
	elif specShader == Material.Shaders['SPEC_PHONG']:
		specFunc = "Blnd_LIPhong(Nf, -normalize(I), hard)"
	elif specShader == Material.Shaders['SPEC_BLINN']:
		specFunc = "Blnd_LIBlinn(Nf, -normalize(I), hard, %.3f)" % mat.refracIndex
	elif specShader == Material.Shaders['SPEC_TOON']:
		specFunc = ('Blnd_LIToonSpecular(Nf, -normalize(I), %.3f, %.3f)' %
			(mat.specularSize, mat.specularSmooth))
	elif specShader == Material.Shaders['SPEC_WARDISO']:
		specFunc = "Blnd_LIWardIso(Nf, -normalize(I), %.3f)" % mat.rms

	writeln(f,'\tCi = Ka * ambient() +')
	writeln(f,'\t\tCol * (Kd * %s +' % diffuseFunc)
	writeln(f,'\t\tKt * %s) +' % transluFunc)
	writeln(f,'\t\tKs * Csp * %s;' % specFunc)
	writeln(f,'\tCi *= Oi;')

writeln(f,'}')

f.close()
