######-----Read .blend-----########
# BLENDFILE READER, ALMOST COMPLETELY ADAPTED FROM NIRVED'S ORIGINAL C CODE
# with quite a few improvements, that makes it even work better than the BLASC script.
# ----------
# speed now very much improved, only the structures requested will be extracted
# Also is able (?) convert the raw data to actual usable python data
#
# Aug 22, 2002 --- Alfredo 'Eeshlo' de Greef
#
# idea kickstarted by post from 'Gregor' on Jandro's YafRay forum,
# who apparently created his own version of something like this. I hope he doesn't mind...
#
# Aug 25, 2002, Now with float, char, int array support (spec. Nurb knots)
# Also better interface to the data, similar to the old Blender210 module.
# The raw output is not anymore what it really looks like in the actual blendfile,
# changes are made while it is parsed, just to make things easier for the python converter.
# So don't rely on the raw output to be the exact contents of the blendfile.
#
# Sep 30, 2002, Added pointer to pointer support (spec. Nurb/MBall materials)
# Oct 1, 2002, Corrected float-, char-, int-, ptr- array length,
#				and so also the actual content of the array, which had too many elements previously.
# Nov 1, 2002, Now empty entries are not skipped any more, this will increase the amount of returned data, but is
#				important if things like material/texture indices are to be used with these lists

from struct import unpack

# still have to deal with endianess, ignored for now
#sys.byteorder=='big'|'little'

class BFReader:

	def __init__(self, filename, getalldefs=0):
		# Gets the initial data strings from the specified file
		
		self.__BVERSION	=	None	# Blender version which was used to save the file
		self.__NAME		=	None	# list of Blender variable names
		self.__TYPE		=	None	# list of Blender variable type names
		# TLEN, which was constructed as a dictionary of variable types and their size, is not used
		#self.__TLEN	=	None
		self.__STRC		=	None
		# all Blender struct definitions as a list of string blocks, only created if requested
		self.STRUCT_DEFLIST	=	[]

		try:
			fp = open(filename, 'rb')
			self.__data = fp.read()
			fp.close()
		except:
			raise "Can't open file"
			return

		#print "\nSTART"

		if not self.__data[:7]=='BLENDER': raise "Not a Blender file???"
		self.__BVERSION = int(self.__data[9:12])
		#print "\nFile was saved with Blender version: %d" % self.__BVERSION

		# get all data blocks as strings
		# TO PREVENT ERRORS WHEN THIS SCRIPT IS USED IN A BLENDFILE WHICH IS CURRENTLY ANALYSED
		# IT NEEDS TO SEARCH FOR THE DATA ID'S BACKWARDS, OTHERWISE THE FIRST HIT IS IN THE SCRIPT ITSELF,
		# WHICH COMPLETELY CONFUSES THE PARSER (even the BLASC script will fail miserably, crashing the system)
		NAME_OK = (self.__data.rfind('SDNANAME')+8, self.__data.rfind('TYPE'))
		TYPE_OK = (self.__data.rfind('TYPE')+4, self.__data.rfind('TLEN'))
		TLEN_OK = (self.__data.rfind('TLEN')+4, self.__data.rfind('STRC'))
		STRC_OK = (self.__data.rfind('STRC')+4, self.__data.rfind('ENDB'))
		if -1 in (NAME_OK + TYPE_OK + TLEN_OK + STRC_OK):
			raise "data error..."

		self.__NAME_dt = self.__data[NAME_OK[0] : NAME_OK[1]]
		self.__TYPE_dt = self.__data[TYPE_OK[0] : TYPE_OK[1]]
		#self.__TLEN_dt = self.__data[TLEN_OK[0] : TLEN_OK[1]]
		self.__STRC_dt = self.__data[STRC_OK[0] : STRC_OK[1]]

		self.__READDATA(getalldefs)


	def __READDATA(self, getalldefs):

		# BLENDER VARIABLE NAMES (NAME)
		names_len = unpack('l', self.__NAME_dt[:4])[0]
		#print 'Number of Blender variables: %d' % names_len
		self.__NAME = [name for name in self.__NAME_dt.split('\0')[1:] if name]

		# TYPE NAMES (TYPE)
		types_len = unpack('l', self.__TYPE_dt[:4])[0]
		#print 'Number of Blender types: %d' % types_len
		self.__TYPE = [name for name in self.__TYPE_dt.split('\0')[1:] if name]

		# NOT USED
		# TYPE SIZES (TLEN)
		# make dictionary containing variable name and size in bytes
		#self.__TLEN = {}
		#for i in range(types_len):
		#	self.__TLEN[self.__TYPE[i]] = unpack('H', self.__TLEN_dt[i*2:i*2+2])[0]

		# list of struct types
		self.__STRUCT_TPLIST = []

		# STRUCTS (STRC)
		strc_len = unpack('l', self.__STRC_dt[:4])[0]
		#print 'Number of Blender structs: %d' % strc_len
		mti = 4
		self.__STRC= []
		for ml in range(strc_len):
			tp, sz = unpack('2H', self.__STRC_dt[mti:mti+4])
			if getalldefs:
				# include all struct definitions from Blendfile
				st = "Nr: %d\nstruct %s {\n" % (ml, self.__TYPE[tp])
				for idx in range(sz):
					p = mti + idx*4 + 4
					t1, t2 = unpack('2H', self.__STRC_dt[p:p+4])
					st += "\t%-20s %s;\n" % (self.__TYPE[t1], self.__NAME[t2])
				st += "};\n\n"
				self.STRUCT_DEFLIST.append(st)
			self.__STRC.append(mti)
			# create list of structs as replacement for FINDSTRUCT(),
			# this way a simple index check is sufficient, which speeds up things by as much as 10 times!
			self.__STRUCT_TPLIST.append(unpack('H', self.__STRC_dt[mti:mti+2])[0])
			mti += (sz*2 + 2)*2


	def getData(self, filterlist=None):
		# Data starts at 12th byte (first 12 is BLENDER_vxxx)
		pt = 12
		# list of ascii datablocks which is returned
		self.__DATA_LIST = []
		name = ''
		self.filterlist = filterlist
		pr_count = 0	# #print info counter, every 1024 'steps', #printing every step slows things down too much
		# first pass through the data to get every struct address and the length of the data at that address
		# Especially DATA/Link are important as these are the places where array data resides (for float*, int*, char* and so on...)
		self.__PTINFO = {}		# pointer info dict, mapping pointer to data length
		self.__LINK_DICT = {}	# Link info dict, mapping pointer to 'local' pointer in __data as well as length of data at address
		#print "\nCollecting addresses..."
		while name!='DNA1':
			#if (pr_count & 31)==0: #print "\r%d / %d" % (pt, len(self.__data)),
			#pr_count += 1
			name = self.__data[pt:pt+4].split('\0')[0]
			pt += 4
			size, pointer, type, number = unpack('4l', self.__data[pt:pt+16])
			# create hexadecimal number as string
			pt += 16
			t = self.__STRC[type]
			key = self.__TYPE[unpack('H', self.__STRC_dt[t:t+2])[0]]
			# store any DATA/Link mapping real pointer to 'local' pointer, also store data block length
			if name=='DATA' and key=='Link':
				self.__LINK_DICT[pointer] = (pt, size)
			if filterlist:
				ADD2DATA = key in filterlist
			else:
				ADD2DATA = 1
			if ADD2DATA:
				self.__PTINFO[pointer] = number
			if name=='REND':
				pt += size
				continue
			pt += size
		#print "\r%d / %d" % (pt, len(self.__data))
		pt = 12
		name = ''
		pr_count = 0
		#print "\nParsing..."
		while name!='DNA1':
			#if (pr_count & 31)==0: #print "\r%d / %d" % (pt, len(self.__data)),
			#pr_count += 1
			name = self.__data[pt:pt+4].split('\0')[0]
			pt += 4
			size, pointer, type, number = unpack('4l', self.__data[pt:pt+16])
			pt += 16
			t = self.__STRC[type]
			key = self.__TYPE[unpack('H', self.__STRC_dt[t:t+2])[0]]
			# if filterlist specified only return data that is in the list,
			# otherwise return ALL data
			if filterlist:
				ADD2DATA = key in filterlist
			else:
				ADD2DATA = 1
			if ADD2DATA:
				# use hex for pointer
				hexpt = "0x%08x" % pointer
				# struct start title
				# order: NAME[SIZE], hexaddress, typenumber, typename, number of structs 
				stblock = "%s[%d] %s %d %s    %d\n" % (name, size, hexpt, type, key, number)
			if name=='REND':
				if ADD2DATA: self.__DATA_LIST.append(stblock)
				pt += size
				continue
			if ADD2DATA:
				# make pt a list to simulate argument by reference for other funcs called from _#print_STRUCT()
				ppt = [pt]
				if number==1:
					stblock += self.__print_STRUCT(ppt, 0, type, 0)
					self.__DATA_LIST.append(stblock)
				else:
					# A list of structs, since this list can be HUGE as a single ascii string, it is still written as separate blocks
					# To indicate a list is coming up, the normal struct info header is used as a single line
					# starting with the keyword LIST, the last number is the number of structs that follow.
					# Each block starts with the hex address,
					# this time using '0m' as prefix, 'm' for multiple, 'l' might look too much like 1,
					# then struct name & the index number
					# The reason for this prefix is that the address might not be the real address,
					# so this is to make sure it doesn't clash with a real address
					hexpt = "0m%08x" % pointer
					stblock = "%s[%d] %s %d %s    %d\n" % (name, size, hexpt, type, key, number)
					self.__DATA_LIST.append("LIST %s" % stblock)
					# each block starts with the hex address (this time using '0m' as prefix), struct name & the index number
					blocksize = size/number
					for i in range(number):
						hexpt = "0m%08x" % (pointer + i*blocksize)
						stblock = ("%s %s --- %d\n" % (hexpt, key, i)) + self.__print_STRUCT(ppt, 0, type, 0)
						self.__DATA_LIST.append(stblock)
			pt += size
		#print "\r%d / %d" % (pt, len(self.__data))
		#print "\nEND\n"
		return self.__DATA_LIST


	def __print_STRUCT(self, pt, parent, number, level):
		# NOTE: _#print_ELEMENT, _#print_STRUCT, _#print_ARRAY always need to be called,
		# regardless if the data is included or not, since the pointer needs to be updated too
		st = self.__STRC[number]
		type, size = unpack('2H', self.__STRC_dt[st:st+4])
		types = self.__TYPE[type]
		st += 4
		stblock = ''
		if parent:
			longname = '%s%s.' % (parent, types)
		else:
			longname = '%s.' % types
		for i in range(size):
			type1, t2 = unpack('2H', self.__STRC_dt[st:st+4])
			typename1, name1 = self.__TYPE[type1], self.__NAME[t2]
			st += 4
			if name1[0]=='*' or name1[0]=='(': #*|**|(*...)()
				# pointer types
				if name1[:2]=='**':
					# pointer to pointer to object, convert to pointer to actual object
					adr = unpack('l', self.__data[pt[0]:pt[0]+4])[0]
					pt[0] += 4
					try:
						locpt, dtsize = self.__LINK_DICT[adr]
					except:
						pass
					else:
						# pointer
						ptr_array = []
						dtsize >>= 2	# dtsize is num. of bytes, convert to num. of pointers
						unpack_str = '%dl' % dtsize
						ptr_array = list(unpack(unpack_str, self.__data[locpt:locpt+(dtsize<<2)]))	# dtsize*sizeof(ptr)
						# make hexpointer list
						ptr_arstr = '['
						for p in ptr_array:
							ptr_arstr += "0x%08x, " % p
						ptr_arstr = ptr_arstr[:-2] + ']'
						# make pointer to pointer a pointer (single *)
						name1 = name1[1:]
						stblock += "    %s \t%s%s[%d] = %s\n" % (typename1, longname, name1, dtsize, ptr_arstr)
				elif name1.find('[')!=-1:
					# array of pointers
					stblock += self.__print_ARRAY(pt, type1, name1, longname, type, types, typename1)
				elif typename1 in ['float', 'int', 'long', 'char']:
					# Possible pointer to DATA/Link here, check if in LINK_DICT
					# Which it should be since all possible Link/DATA address blocks are stored
					adr = unpack('l', self.__data[pt[0]:pt[0]+4])[0]
					pt[0] += 4
					try:
						locpt, dtsize = self.__LINK_DICT[adr]
					except:
						pass
					else:
						# DATA/Link block found, get the data at that address
						# since the data is now written directly with the variable,
						# it is thus treated as not being a pointer anymore, so remove * from name
						name1 = name1[1:]
						if typename1=='float':
							# pointer to floating point array
							fp_array = []
							dtsize >>= 2	# dtsize is num. of bytes, convert to num. of floats
							unpack_str = '%df' % dtsize
							fp_array = list(unpack(unpack_str, self.__data[locpt:locpt+(dtsize<<2)]))	# dtsize*sizeof(float)
							stblock += "    %s \t%s%s = %s\n" % (typename1, longname, name1, fp_array)
						elif typename1=='int' or typename1=='long':
							# pointer to integer array
							int_array = []
							dtsize >>= 2	# dtsize is num. of bytes, convert to num. of ints
							unpack_str = '%dl' % dtsize
							int_array = list(unpack(unpack_str, self.__data[locpt:locpt+(dtsize<<2)]))	# dtsize*sizeof(int|long)
							stblock += "    %s \t%s%s = %s\n" % (typename1, longname, name1, int_array)
						elif typename1=='char':
							# pointer to array of char, a string, dtsize already is in bytes
							stblock += "    %s \t%s%s = " % (typename1, longname, name1)
							# use string as is, using `, up to first \0
							stblock += `self.__data[locpt:locpt+dtsize-1].split('\0')[0]` + '\n'
				else:
					# test if it is in the PTINFO list, and if it is and the number of structs at that address is more than 1,
					# write the address as '0m' to make sure that it coincides with the list written at some point later or earlier.
					# this is all done for the python conversion.
					adr = unpack('l', self.__data[pt[0]:pt[0]+4])[0]
					hexpt = "0x%08x" % adr
					pt[0] += 4
					try:
						lnum = self.__PTINFO[adr]
					except:
						pass
					else:
						if lnum>1: hexpt = '0m' + hexpt[2:]
					stblock += "    %s \t%s%s = %s\n" % (typename1, longname, name1, hexpt)
			else:
				try:
					# struct?
					fst = self.__STRUCT_TPLIST.index(type1)
				except:
					if name1.find('[')!=-1:
						# array
						stblock += self.__print_ARRAY(pt, type1, name1, longname, type, types, typename1)
					else:
						# element
						el = self.__print_ELEMENT(pt, 0, type1, name1, longname, type, types, typename1)
						stblock += "    %s \t%s%s = %s\n" % (typename1, longname, name1, el)
				else:
					# struct, test if array
					isarray = name1.find('[')
					if isarray!=-1:
						# Add struct array elements as a separate struct block outside of main struct,
						# needs to be done otherwise the data can't be correctly converted to python data.
						# Note that this is the only other case (see pointer above) that the struct name does not need to be
						# included in the filterlist
						# The array is written as if it were a pointer to the array.
						dim = int(name1[isarray+1:].split(']')[0])
						ptlist = "["	# list of hex numbers, must be unquoted values, cannot do '%s' % list
						lev = level+1
						for i in range(dim):
							# since we are not using an actual address from the file but the index into the data instead,
							# the numbers are prefixed with '0p' instead of '0x' to prevent clashes with real existing addresses
							hexpt = "0p%08x" % pt[0]
							ptlist += hexpt
							if i!=(dim-1): ptlist += ', '
							subs = "# %s # %s #\n" % (hexpt, typename1)
							subs += self.__print_STRUCT(pt, 0, fst, lev)
							self.__DATA_LIST.append(subs)
						ptlist += "]"
						# Note the * in next line string, write as if pointer to array data
						stblock += "    %s \t%s*%s = %s\n" % (typename1, longname, name1, ptlist)
					else:
						# Once again a change to the real raw data for the python conversion.
						# struct variable found, write out as separate block, and write variable as pointer to it
						# Since again the 'pointer' into the string data is used as the address,
						# the prefix this time is '0s' (s for sub)
						hexpt = '0s%08x' % pt[0]
						# '*' in line below to make name a pointer similar to struct array above
						stblock += '    %s \t%s*%s = %s\n' % (typename1, longname, name1, hexpt)

						sb_st = self.__STRC[fst]
						sb_type, sb_size = unpack('2H', self.__STRC_dt[sb_st:sb_st+4])
						sb_types = self.__TYPE[sb_type]

						subblock = "%s[%d] %s %d %s    1\n" % (sb_types, sb_size, hexpt, sb_type, sb_types)
						subblock += self.__print_STRUCT(pt, '', fst, level+1)
						self.__DATA_LIST.append(subblock)
		return stblock


	def __print_ARRAY(self, pt, type1, name1, longname, type, typs, typename1):
		# Arrays are written as a list, list of lists if two dimensional.
		# Nothing is done to make it look good, elements are just appended,
		# this is just done to make it easy to parse with python directly.
		dim2 = 0
		c1 = name1.find('[')
		tst = name1[c1+1:].replace('[', ' ').replace(']', ' ').split()
		dim1, dim2 = int(tst[0]), None
		if len(tst)>1: dim2 = int(tst[1])
		buf = name1[:c1]
		ispointer = (name1.find('*')!=-1)
		st = ''
		if dim2:
			# two dimensional array
			st = "    %s \t%s%s[%d][%d] = [" % (typename1, longname, buf, dim1, dim2)
			for i in range(dim1):
				st += '['
				for j in range(dim2):
					st += self.__print_ELEMENT(pt, ispointer, type1, name1, longname, type, typs, typename1)
					if j!=(dim2-1): st += ', '
				st += ']'
				if i!=(dim1-1): st += ', '
			st += ']\n'
		else:
			# one dimensional array
			if typename1=='char':
				st += "    %s \t%s%s[%d] = " % (typename1, longname, buf, dim1)
				# note the ` in next line, necessary to make sure the string is included as is, no escape char weirdness
				# also make sure only the string up to the first zero delimiter char is used
				st += `self.__data[pt[0]:pt[0]+dim1].split('\0')[0]` + '\n'
				pt[0] += dim1
			else:
				st = "    %s \t%s%s[%d] = [" % (typename1, longname, buf, dim1)
				for i in range(dim1):
					st += self.__print_ELEMENT(pt, ispointer, type1, name1, longname, type, typs, typename1)
					if i!=(dim1-1): st += ', '
				st += ']\n'
		return st


	def __print_ELEMENT(self, pt, ispointer, type1, name1, longname, type, typs, typename1):
		typeS = self.__TYPE[type1]
		st = ''
		if ispointer:
			st = "0x%08x" % unpack('l', self.__data[pt[0]:pt[0]+4])
			pt[0] += 4
		elif typeS=='char':
			st = "%d" % unpack('b', self.__data[pt[0]])
			pt[0] += 1
		elif typeS=='uchar':
			st = "%d" % unpack('B', self.__data[pt[0]])
			pt[0] += 1
		elif typeS=='short':
			st = "%d" % unpack('h', self.__data[pt[0]:pt[0]+2])
			pt[0] += 2
		elif typeS=='ushort':
			st = "%d" % unpack('H', self.__data[pt[0]:pt[0]+2])
			pt[0] += 2
		elif typeS=='int' or typeS=='long':
			st = "%d" % unpack('l', self.__data[pt[0]:pt[0]+4])
			pt[0] += 4
		elif typeS=='ulong':
			st = "%d" % unpack('L', self.__data[pt[0]:pt[0]+4])
			pt[0] += 4
		elif typeS=='float':
			st = "%f" % unpack('f', self.__data[pt[0]:pt[0]+4])
			pt[0] += 4
		elif typeS=='double':
			st = "%f" % unpack('d', self.__data[pt[0]:pt[0]+8])
			pt[0] += 8
		else:
			st = "?TYPE? %d %s" % (type1, typeS)
		return st


	# Using the unique abilities of python, the raw ascii data
	# is converted to a set of classes which is then filled in with new data members.
	# These automatically have the correct types.
	# This way it is all completely blendfile version indepent.
	def convert2Python(self):
		# list of returned classes with data
		Clist = []
		# Cdatamap -> list of [Clist index, datamember1, datamembern...] lists
		Cdatamap = []
		# Cadrsmap -> dict of 'address: Clist index'
		Cadrsmap = {}
		# Clistadrs -> list of dictionairies of list element pointers
		Clistadrs = []
		curidx = 0		# current index in Clist
		ListLength = 0	# list counter for struct lists
		for block in self.__DATA_LIST:

			lines = block.splitlines()

			# first line contains struct name, address etc
			names = lines[0].split()

			# if a list follows, the first line starts with LIST
			# only get new address & name when list read is not completed yet
			if ListLength==0:
				if names[0]=="LIST":
					elemp = {}
					address, BClass, ListLength = names[2], names[4], int(names[5])
					continue
				else:
					ListLength = 0
					address, BClass = names[1], names[3]

			# decrease ListLength to use it as a counter for the elements and get the address of this element
			if ListLength!=0:
				address = names[0]
				# add current Clist index to element list
				elemp[address] = curidx
				ListLength -= 1
				if ListLength==0:
					# end of list reached, include in Clistadrs
					Clistadrs.append(elemp)
	
			# Create a new class (if it doesn't exist yet), the only methods added are for init & repr.
			# __init_ creates three variables, id, name & address. id is the blender ID struct which contains the name as used in Blender
			# but since sometimes (like for MTex or EnvMap) id or name might not be defined, the address will be used instead.
			# address is a hexadecimal string (possibly starting with 0p/0s/0m instead of 0x, see #print_STRUCT)
			st = "try:\n"
			st += "\tdmy = " + BClass + "()\n"
			st += "except:\n"
			st += "\tclass " + BClass + ":\n"
			st += "\t\tdef __init__(self):\n"
			st += "\t\t\tself.id\t=\tNone\n"
			st += "\t\t\tself.name\t=\tNone\n"
			st += "\t\t\tself.address\t=\tNone\n"
			st += "\t\tdef __repr__(self):\n"
			st += "\t\t\tif self.id:\n"
			st += "\t\t\t\treturn '[" + BClass + " %s]' % self.id.name\n"
			st += "\t\t\telif self.name:\n"
			st += "\t\t\t\treturn '[" + BClass + " %s]' % self.name\n"
			st += "\t\t\treturn '[" + BClass + " %s]' % self.address\n"
			st += "\tdmy = " + BClass + "()\n"
			exec(st)

			dmap = [curidx]

			# first line can be skipped in any case, already processed above
			for line in lines[1:]:
				# split first at '=' (only first occurence, might be string with equal sign as char data comeing after it)
				nline = line.split('=', 1)
				# note Type, not type, function type() is used below
				Type, varname = nline[0].split()[:2]
				value = nline[1]
				# if there is anything more than one dot used (before '=', not counting fpnums of course), ignore, unless it is ID.name
				if varname.count('.')>1:
					continue
				varname = varname.split('.')[1]	# only interested in the actual member name, already know struct
				# define as class data member, unless it is a pointer or pointer array
				if varname[:2]=='**' or varname[:2]=='(*':
					# ignore pointer to pointer or pointer to function
					continue
				elif varname[0]=='*':
					# pointer, use address or address list, will be replaced by actual class instance later, if possible
					varname = varname[1:]
					# if array, remove []
					isarray = varname.find('[')
					if isarray!=-1:
						# array, create address list
						varname = varname[:isarray]
						dmap.append(varname)
						# previously this part ignored empty entries in the list,
						# but that causes problems when trying to use list indices that are meant
						# to be used with this like material and texture lists
						vlist = [v for v in value[2:len(value)-1].replace(',', '').split()]
						exec('dmy.' + varname + '=vlist')
					else:
						# single pointer
						dmap.append(varname)
						if varname=='from':
							# I'm not sure where it is used, but 'from' seems to be a name used somewhere in python (class/module?)
							# it generates an error, so change it to 'From' instead
							varname = 'From'
						v = value.strip()
						exec('dmy.' + varname + '=v')
				else:
					# if array, remove [xx]
					isarray = varname.find('[')
					if isarray!=-1: varname = varname[:isarray]
					exec('dmy.' + varname + '=' + value)
			# set the address in any case, since IDname or name might not be defined
			dmy.address = address
			Clist.append(dmy)
			Cdatamap.append(dmap)
			Cadrsmap[address] = curidx
			curidx += 1
		# replace found pointers with the actual class instances if posssible, otherwise replace by None
		for CL in Cdatamap:
			CLS_inst = Clist[CL[0]]
			for datamember in CL[1:]:
				if datamember=='from': datamember = 'From'	# see above
				exec('pointers = CLS_inst.' + datamember)
				if type(pointers).__name__=='list':
					# list
					i = 0
					for p in pointers:
						try:
							actual_inst = Clist[Cadrsmap[p]]
						except:
							actual_inst = None
						exec('CLS_inst.' + datamember + '[i]=actual_inst')
						i += 1
				elif pointers[:2]=='0m':
					# pointer to list, collect all elements as a single list if possible
					idx = 0
					for elemp in Clistadrs:
						if pointers in elemp.keys(): break
						idx += 1
					if idx<len(Clistadrs):
						# pointer found, create list of instances
						exec('CLS_inst.' + datamember + '=[]')
						elemp = Clistadrs[idx]
						for i in range(len(elemp)):
							actual_inst = Clist[elemp.values()[i]]
							exec('CLS_inst.' + datamember + '.append(actual_inst)')
					else:
						# pointer not found, use None
						exec('CLS_inst.' + datamember + '=None')
				else:
					try:
						actual_inst = Clist[Cadrsmap[pointers]]
					except:
						actual_inst = None
					exec('CLS_inst.' + datamember + '=actual_inst')

		# all data collected, return BlenderF instance
		return BlenderF(Clist)




# Simple class to make it even easier to get the useful data
# Uses similar method names like the old Blender210 module
# For example:
# 	material = BlenderF.getMaterial('Red')
# Note that not all data is supported yet, it also does not necessarily return the full data you might need
# Some of them might even be useless, to get any other type, use getOther(TYPE, NAME)

class BlenderF:

	def __init__(self, Clist):
		self.Clist = Clist
		self.meshes = []

	def __search(self, Type, NAME=''):
		# checks for name plus and without ID chars (MA/TX/IM..etc..)
		# if NAME not given, returns a list of all instances of type Type
		if NAME!='':
			for C in self.Clist:
				tp = repr(C)[1:-1].split()[0]
				try:
					nm = C.id.name
				except:
					nm = C.name
				if tp==Type and (nm==NAME or nm[2:]==NAME):
					return C
		else:
			ilist = []

			for C in self.Clist:
				tp = repr(C)[1:-1].split()[0]
                                if tp == Type: ilist.append(C)
			return ilist
			

	def getObject(self, NAME=''):
		"Returns Object"
		return self.__search('Object', NAME)
	    
	def getProperty(self, NAME=''):
		"Returns Property"
		return self.__search('bProperty', NAME)
	    
	def getMesh(self, NAME=''):
		"Returns Mesh"
		return self.__search('Mesh', NAME)

	def getMaterial(self, NAME=''):
		"Returns Material"
		return self.__search('Material', NAME)

	def getMTex(self, NAME=''):
		"Returns Material Texture"
		return self.__search('MTex', NAME)

	def getTex(self, NAME=''):
		"Returns Texture"
		return self.__search('Tex', NAME)

	def getColorBand(self, NAME=''):
		"Return ColorBand"
		return self.__search('ColorBand', NAME)

	def getEnvMap(self, NAME=''):
		"Returns Environmentmap"
		return self.__search('EnvMap', NAME)

	def getImage(self, NAME=''):
		"Returns Image"
		return self.__search('Image', NAME)

	def getNurb(self, NAME=''):
		"Returns NURB"
		return self.__search('Nurb', NAME)

	def getIpo(self, NAME=''):
		"Returns IPO"
		return self.__search('Ipo', NAME)

	def getCamera(self, NAME=''):
		"Returns Camera"
		return self.__search('Camera', NAME)

	def getMetaBall(self, NAME=''):
		"Returns MetaBall"
		return self.__search('MetaBall', NAME)

	def getMetaElement(self, NAME=''):
		"Returns MetaElement"
		return self.__search('MetaElement', NAME)

	def getCurve(self, NAME=''):
		"Returns Curve"
		return self.__search('Curve', NAME)

	def getLamp(self, NAME=''):
		"Returns Lamp"
		return self.__search('Lamp', NAME)

	def getPluginTex(self, NAME=''):
		"Returns Texture Plugin info"
		return self.__search('PluginTex', NAME)

	def getPluginSeq(self, NAME=''):
		"Returns Sequence Plugin info"
		return self.__search('PluginSeq', NAME)

	def getLattice(self, NAME=''):
		"Returns Deform Lattice"
		return self.__search('Lattice', NAME)

	def getWorld(self, NAME=''):
		"Returns World"
		return self.__search('World', NAME)

	def getRenderData(self, NAME=''):
		"Returns RenderData"
		return self.__search('RenderData', NAME)

	def getPartEff(self, NAME=''):
		"Returns Particle Effect System"
		return self.__search('PartEff', NAME)

	def getWaveEff(self, NAME=''):
		"Returns Wave Effect"
		return self.__search('WaveEff', NAME)

	def getBuildEff(self, NAME=''):
		"Returns Build effect"
		return self.__search('BuildEff', NAME)

	def getBone(self, NAME=''):
		"Returns Bone"
		return self.__search('Bone', NAME)

	def getOther(self, TYPE, NAME=''):
		"Returns obect of type 'TYPE' and name 'NAME'"
		return self.__search(TYPE, NAME)



##########-----Read Nurbs ------#######
#from BFREAD import *
class object: pass

class nObject:
    def __init__(self):
        self.vecs = []
        self.address = []
        self.nu = 0
        self.uorder = 0
        self.uknot = 0
        self.umin = 0
        self.umax = 0
        self.nv = 0
        self.vorder = 0
        self.vknot = 0
        self.vmin = 0
        self.vmax = 0


def adr_cmp(a, b):
   x = '0x'+a.address[2:]
   y = '0x'+b.address[2:]
   exec('t = %s-%s' % (x,y))
   return t

def propLoop(var):
    output = ""
    for val in dir(var):
        output += "\t"+val, eval("var."+val)+"\n"
    return output
def load(file):
    bf = BFReader("C:/Documents and Settings/Administrator/Desktop/Client Work/RendermanGUI/nurbTest2.blend")
    bf.getData()
    blend = bf.convert2Python()
    objs = blend.getObject()
    nurbs = blend.getNurb()

    nobjs = []
    objects = []
    
    for nurb in nurbs:
        nObj = nObject()
        nObj.address = nurb.address
        #print len(nurb.bp)
        for bp in nurb.bp:
            nObj.vecs.append(bp.vec)
        nobjs.append(nObj)

    for obj in objs:
        OB              = object()
        if str(obj)[10:-1].find(".") > 0:
            empty,name = "",str(obj)[10:-1]
            dName =  name.split(".")
            for i in dName: empty += i
            OB.name         = empty
        else:
            #print obj.id
            OB.name     = str(obj)[10:-1]
        OB.orientation  = obj.obmat
        OB.position     = obj.loc
        OB.origin       = obj.orig
        OB.parent       = obj.parent
        OB.nurb         = []

        for data in dir(obj):
            s2 = eval("obj."+str(data))
            #print s2
            if str(data) == "data":
                if str(s2).find("Curve") >= 0:
                    print obj.id
                    first = s2.nurb.first
                    #print  first.pntsu, first.orderu, first.knotsu, first.resolu
                    for nobj in nobjs:
                        if nobj.address == s2.nurb.first.address:
                            nobj.nu = first.pntsu
                            nobj.uorder = first.orderu
                            nobj.uknot = first.knotsu
                            nobj.umin = first.resolu
                            nobj.umax = first.resolu
                            nobj.nv = first.pntsv
                            nobj.vorder = first.orderv
                            nobj.vknot = first.knotsv
                            nobj.vmin = first.resolv
                            nobj.vmax = first.resolv
                            OB.nurb.append(nobj)
                    for val in dir(s2.nurb.first):
                        if val == "next":
                            next = eval("s2.nurb.first."+val)
                            while 1:
                                try:
                                    for nobj in nobjs:
                                        #print nobj.address, s2.nurb.first.address
                                        if nobj.address == next.address:
                                            nobj.nu = next.pntsu
                                            nobj.uorder = next.orderu
                                            nobj.uknot = next.knotsu
                                            nobj.umin = next.resolu
                                            nobj.umax = next.resolu
                                            nobj.nv = next.pntsv
                                            nobj.vorder = next.orderv
                                            nobj.vknot = next.knotsv
                                            nobj.vmin = next.resolv
                                            nobj.vmax = next.resolv
                                            OB.nurb.append(nobj)
                                    next = next.next
                                except: break
                    objects.append(OB)

    print "Load Complete"
    return objects


#########-----Main-######################
import Blender, pickle, zlib, time
from Blender import Draw, BGL, Window
from Blender.BGL import *
from math import pi



render = Blender.Scene.GetCurrent().getRenderingContext()

class Scene:
	def __init__(self):
		self.animated = False
		self.start = render.startFrame()
		self.end = render.endFrame()
		self.output = "file"
		self.resolution = (render.imageSizeX(),render.imageSizeY())
		self.pixelSamples = 1
		self.shaderRate = 1
		self.camera = Camera()
		self.lights = []
		self.meshes = []
		self.nurbs = []
		self.curves = []
		self.material = []
	def addMesh(self,mesh):
		self.meshes.append(mesh)
	def addCurve(self,curve):
		self.curves.append(curve)
	def addLight(self,light):
		self.lights.append(light)
class Camera:
	def __init__(self):
		self.name 	= ""
		self.start 	= []
		self.end 	= []
		self.lens 	= []
		self.trans	= []
	def addTrans(self,trans):
		self.trans.append(trans)
	def addLens(self,lens):
		self.lens.append(lens)
	def addClip(self,end,start):
		self.start.append(start)
		self.end.append(end)
class Mesh:
	def __init__(self):
		self.name = ""
		self.type = "PointPolygons"
		self.mat = None
		self.IPO = []
		self.trans = []
		self.geo = None
		self.bounds = []
class geometry:
	length	= None
	nvert 	= None
	vert 	= None
	color 	= None
	normal 	= None
	alpha 	= None
class Curve:
	def __init__(self):
		self.name = ""
		self.data = ""
		self.width = []
		self.mat = None
		self.trans = []
		self.curves= []
		self.bounds = []
class Lamp:
	def __init__(self):
		self.name = ""
		self.shadowname = ""
		self.coneangle,self.conedeltaangle = (0,0)
		self.color = (0,0,0) 
		self.energy = 0.0
		self.samples = 1
		self.blur = 0.0
		self.buff = 512
		self.trans = []
		self.lock = []
		self.look = []
	def setLook(self,x,y,z,toY,toX,toZ):
		self.lock.append([x,y,z])
		self.look.append([toY,toX,toZ])

def getBounds(obj):
	x = []
	y = []
	z = []
	for pos in obj.getBoundBox():
		(bx,by,bz) = pos
		x.append(bx)
		y.append(by)
		z.append(bz)
	
	x.sort()
	y.sort()
	z.sort()

	boundingbox =[x[0],x[-1],y[0],y[-1],z[0],z[-1]]
	
	return boundingbox
#gather object mesh data
def getMesh(name,mClass):
	mesh = Blender.NMesh.GetRawFromObject(name)
	object = Blender.Object.Get(name)
	try:
		mClass.image = object.getData().faces[0].image.filename
	except: pass
	mClass.name = object.getName()
	mClass.trans.append(writeMatrix(mClass.name))
	mClass.geo = getGEO(name)
	mClass.bounds = getBounds(object)
#gather object curve data
def getCurve(name,mClass):
	object = Blender.Object.Get(name)
	mClass.name = object.getName()
	mClass.trans.append(writeMatrix(mClass.name))
	mClass.bounds = getBounds(object)
#gather lamp data
def getLamp(name,lClass):
	lampobj = Blender.Object.Get(name)
	lamp = Blender.Lamp.Get(lampobj.getData().name)
	x = lampobj.matrix[3][0] / lampobj.matrix[3][3]
	y = lampobj.matrix[3][1] / lampobj.matrix[3][3]
	z = lampobj.matrix[3][2] / lampobj.matrix[3][3]
	toX = -lampobj.matrix[2][0] + lampobj.matrix[3][0]
	toY = -lampobj.matrix[2][1] + lampobj.matrix[3][1]
	toZ = -lampobj.matrix[2][2] + lampobj.matrix[3][2]
	lClass.name = name
	lClass.setLook(x,y,z,toX,toY,toZ)
	lClass.coneangle = lamp.spotSize * pi / 360
	lClass.conedeltaangle = lamp.spotBlend * (lamp.spotSize * pi / 360)
	lClass.energy = lamp.energy
	lClass.color = (lamp.R, lamp.G, lamp.B)
	lClass.buff = lamp.getBufferSize()
	lClass.trans = [writeLampMatrix(lampobj)]
def getGEO(name):
	obj = Blender.Object.Get(name).getData()
	
	vIndex 	= []
	nIndex 	= []
	fIndex 	= []
	fpolys 	= []
	fnorm 	= []
	ecrease	= [[],[],[]]		
	class geometry:
		length	= None
		nvert 	= None
		vert 	= None
		color 	= None
		normal 	= None
		alpha 	= None
		crease	= None

	for vert in obj.verts:
		(nx,ny,nz) = vert.no
		(x,y,z) = vert.co
		vIndex.append([x,y,z])
		nIndex.append([nx,ny,nz])

	for face in obj.faces:
		fpolys.append(len(face))
		for v in face.v:
			fIndex.append(v.index)
			
	if obj.edges:
		ecrease[2] = [10]
		for edge in obj.edges:
			if edge.crease/255.0 == 1.0:
				ecrease[1].append(edge.v1.index)
				ecrease[1].append(edge.v2.index)
		
		ecrease[0] = [len(ecrease[1]),1] 
	
	
	
	
	mesh = geometry()
	mesh.length = fpolys
	mesh.nvert 	= fIndex
	mesh.vert 	= vIndex
	mesh.normal	= nIndex
	mesh.crease	= ecrease
	return mesh
def CameraMat():
	Blender.Get('curframe')
	Blender.Window.RedrawAll()
	cam = Blender.Scene.GetCurrent().getCurrentCamera()
	caminvmatrix = cam.getInverseMatrix()
	bcam = Blender.Camera.Get(cam.getName())
	Scene.camera.addLens(bcam.getLens())
	Scene.camera.addClip(bcam.getClipEnd(), bcam.getClipStart())
	Scene.camera.addTrans("Transform [" +
			"%s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s]\n" %
			(caminvmatrix[0][0],
			caminvmatrix[0][1],
			-caminvmatrix[0][2],
			caminvmatrix[0][3],
			caminvmatrix[1][0],
			caminvmatrix[1][1],
			-caminvmatrix[1][2],
			caminvmatrix[1][3],
			caminvmatrix[2][0],
			caminvmatrix[2][1],
			-caminvmatrix[2][2],
			caminvmatrix[2][3],
			caminvmatrix[3][0],
			caminvmatrix[3][1],
			-caminvmatrix[3][2],
			caminvmatrix[3][3]))
def writeLampMatrix(lamp):
	caminvmatrix = lamp.getInverseMatrix()
	trans = "Transform [" + "%s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s]\n" %(caminvmatrix[0][0],
			caminvmatrix[0][1],
			-caminvmatrix[0][2],
			caminvmatrix[0][3],
			caminvmatrix[1][0],
			caminvmatrix[1][1],
			-caminvmatrix[1][2],
			caminvmatrix[1][3],
			caminvmatrix[2][0],
			caminvmatrix[2][1],
			-caminvmatrix[2][2],
			caminvmatrix[2][3],
			caminvmatrix[3][0],
			caminvmatrix[3][1],
			-caminvmatrix[3][2],
			caminvmatrix[3][3])
	return trans
def writeMatrix(name):
	trans = ""
	Blender.Get('curframe')
	Blender.Window.RedrawAll()
	matrix = Blender.Object.Get(name).matrix
	trans += "\t\tTransform [ %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s]\n"%(matrix[0][0], matrix[0][1],
		matrix[0][2], matrix[0][3],
		matrix[1][0], matrix[1][1],
		matrix[1][2], matrix[1][3],
		matrix[2][0], matrix[2][1],
		matrix[2][2], matrix[2][3],
		matrix[3][0], matrix[3][1],
		matrix[3][2], matrix[3][3])
	return trans

def getObjects(file):
	print file
	if file:
		objs = load(file)
		for obj in objs:
			print dir(obj)
	for object in Blender.Object.Get():
		obName =  object.getName()
		print obName, object.getType()
		if object.getType() == "Mesh":
			
			mesh = Mesh()	
			getMesh(object.getName(),mesh)
			Scene.addMesh(mesh)
		if object.getType() == "Lamp":
			exsits = False
			#check if the lamp is already created
			for lamp in Scene.lights:
				if lamp.name == obName:
					exsits = True
					getLamp(object.getName(),lamp)
			#if not create add a new one
			if exsits == False:
				lamp = Lamp()	
				getLamp(object.getName(),lamp)
				Scene.addLight(lamp)
		if object.getType() == "Surf":
			print "Nurb"
		if object.getType() == "Curve":
			 oCur = Curve()
			 curveString = ""
			 dat = object.getData()
			 if dat.isNurb():
				curve = []
				for c in range(0,dat.getNumCurves()):
					points = []
					curveString += 'Basis "bezier" 3 "bezier" 3\nCurves "cubic" [%s] "nonperiodic" "P" [ '%(dat.getNumPoints())
					for p in range(0,dat.getNumPoints()):
						cord = dat.getControlPoint(c,p)
						curveString += "%s %s %s "%(cord[0],cord[1],cord[2])
						points.append([cord[0],cord[1],cord[2]])
					curveString += '] \n\t\t\t"width" [0.005 0.01]'
					curve.append(points)
				oCur.data = curveString
				oCur.curves.append(curve)
				getCurve(object.getName(),oCur)
				Scene.addCurve(oCur)
			 else:
				 dat = object.getData()
				 curve = []
				 for cur in object.getData():					
					curveString += 'Basis "bezier" 3 "bezier" 3\nCurves "cubic" [%s] "nonperiodic" "P" [ '%(len(cur)*3)
				   	for point in cur:
						points = []
						pos = str(point).split()
						pos = pos[1:10]
						for i in range(0,len(pos),3):
							curveString += "%s %s %s "%(pos[i],pos[i+1],pos[i+2])
							points.append([pos[i],pos[i+1],pos[i+2]])
					curve.append(points)
					curveString += '] \n\t\t\t"width" [0.005 0.01]'
		
				 oCur.data = curveString
				 oCur.curves.append(curve)
				 getCurve(object.getName(),oCur)
				 Scene.addCurve(oCur)




Scene = Scene()	

#Created with The_Nerd's GUI creator V2.400000
EVENT_fileTxt = 10
EVENT_filename = 11
EVENT_anim = 12
EVENT_xport = 13
EVENT_choose_file = 14
EVENT_choose_blend = 15
Object_fileTxt = Draw.Create("")
Object_filename = Draw.Create("")
Object_blendname = Draw.Create("")
Object_blendtxt = Draw.Create("")
Object_anim = Draw.Create(0)
Object_xport = Draw.Create(0)

def draw():
   global EVENT_fileTxt
   global EVENT_filename
   global EVENT_anim
   global EVENT_xport
   global Object_fileTxt
   global Object_filename
   global Object_anim
   global Object_xport
   global EVENT_choose_file,EVENT_choose_blend, Object_blendname

   BGL.glClearColor(0.500000, 0.500000, 0.500000, 0.0)
   BGL.glClear(GL_COLOR_BUFFER_BIT)

   BGL.glColor3f(0.000000, 0.000000, 0.000000)
   BGL.glRasterPos2i(168, 264)
   Object_fileTxt = Draw.Text("Filename:")
   BGL.glRasterPos2i(160, 294)
   Object_blendtxt = Draw.Text("Blend File:")
   Object_filename = Draw.String("", EVENT_filename, 224, 256, 150, 20,Object_filename.val, 1024, "path and file to export too,.")
   Object_blendname = Draw.String("", EVENT_choose_blend, 224, 286, 150, 20,Object_blendname.val, 1024, "Location of blend file")
   Object_anim = Draw.Toggle("Animated", EVENT_anim, 168, 232, 60, 20, Object_anim.val, "If you want to export a scene as animation")
   Object_xport = Draw.Button("Export", EVENT_xport, 312, 232, 60, 20, "Export to sMan")
   Draw.Button("File", EVENT_choose_file, 380, 256, 40, 20)
   Draw.Button("Blend", EVENT_choose_blend, 380, 286, 40, 20)

def event(event, value):
   if event == Draw.ESCKEY and not value: Draw.Exit()

def b_event(event):
   global EVENT_fileTxt
   global EVENT_filename
   global EVENT_anim
   global EVENT_xport
   global Object_fileTxt
   global Object_filename
   global Object_anim
   global Object_xport
   global EVENT_choose_file,EVENT_choose_blend, Object_blendname

   if event == 0: pass
   elif event == EVENT_fileTxt:
      #***Place your code for Object fileTxt here***#
      pass
   elif event==EVENT_choose_file:
      Blender.Window.FileSelector(select_file,"Choose file")
   elif event==EVENT_choose_blend:
      Blender.Window.FileSelector(select_blend,"Choose .blend")
   elif event == EVENT_filename:
      pass
   elif event == EVENT_anim:
      if Object_anim.val: Scene.animated = True
      else: Scene.animated = True
   elif event == EVENT_xport: 
		exportTime = time.time()
		blend = Object_blendname.val
		print "Xporting..."		
		if Scene.animated:
			for frame in range(Scene.start,Scene.end):
				Blender.Set("curframe", frame)
				CameraMat()
				if Object_blendname.val:
					blendFile = Object_blendname.val.replace("\\","/")
					getObjects(blendFile)
				else: getObjects("")
				Blender.Redraw()
		else:
			CameraMat()
			if Object_blendname.val:
				blendFile = Object_blendname.val.replace("\\","/")
				getObjects(blendFile)
			else: getObjects("")
			print "Raw Export complete"
		file = open(Object_filename.val+".sman", "wb")	
		print "Compressing"
		compressed = zlib.compress(pickle.dumps(Scene),zlib.Z_BEST_COMPRESSION)
		file.write(compressed)
		print "Done"
		print "Export time: %s seconds", time.time()-exportTime
		Draw.Exit()
   Draw.Draw()
	
def select_file(filename):
    global Object_filename
    #debut
    Object_filename.val=filename
def select_blend(blendname):
    global Object_blendname
    #debut
    Object_blendname.val=blendname

Draw.Register(draw, event, b_event) 




